package org.netkernel.module.standard;

import java.util.ArrayList;
import java.util.List;

import org.netkernel.container.IKernel;
import org.netkernel.container.ILogger;
import org.netkernel.layer0.meta.IEndpointStateMeta;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.impl.SpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.nkf.impl.*;
import org.netkernel.layer0.util.GoldenThreadExpiryFunction;
import org.netkernel.layer0.util.TupleList;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLUtils;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestResponseFields;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestResponseFieldsImpl;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.urii.ClassLoaderWithExports;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IIdentifier;
import org.netkernel.urii.INetKernelException;
import org.w3c.dom.Element;

public class StandardSpace extends NKFCombinedEndpointSpaceImpl
{
	private final StandardSpace mParentSpace;
	private final StandardModule mOwner;
	private final Element mConfig;
	private final RootSpace mRootSpace;
	private final IRequestScopeLevel mStaticScope;
	private final StandardSpace[] mStaticScopeArray;
	private INKFResponseReadOnly mLastMetaResponse;
	private boolean mIsBinding;
	private boolean mIsBound;
	private boolean mAreEndpointsComissioned;
	private String mName;
	protected IStandardEndpoint[] mMappings;
	private IRequestResponseFields mProblems=RequestResponseFieldsImpl.EMPTY;
	private int mHash;
	private GoldenThreadExpiryFunction mUnboundExpiry;
	
	public StandardSpace(StandardSpace aParentSpace, StandardModule aOwner, Element aConfig)
	{
		declareThreadSafe();
		mParentSpace=aParentSpace;
		mOwner=aOwner;
		mConfig=aConfig;
		mHash=System.identityHashCode(this);
		if (mParentSpace!=null)
		{	StandardSpace root=mParentSpace;
			while(root.getParent()!=null)
			{	root=root.getParent();
			}
			mRootSpace=(RootSpace)root;
		}
		else
		{	mRootSpace=(RootSpace)this;
		}
		mName="";
		
		// build static scope of space
		List<StandardSpace> spaces = new ArrayList<StandardSpace>();
		StandardSpace space = this;
		while (space!=null)
		{	spaces.add(space);
			space=space.getParent();
		}
		IRequestScopeLevel scope = RequestScopeLevelImpl.createRootScopeLevel(spaces.get(spaces.size()-1));
		mStaticScopeArray=new StandardSpace[spaces.size()];
		int i=spaces.size()-1;
		mStaticScopeArray[i]=spaces.get(i);
		for (i=spaces.size()-2; i>=0; i--)
		{	space = spaces.get(i);
			scope = RequestScopeLevelImpl.createNonDurableScopeLevel(space, scope);
			mStaticScopeArray[i]=space;
		}
		mStaticScope=scope;		
	}
	
	public int hashCode()
	{	return mHash;
	}
	
	public void setName(String aName)
	{	mName=aName;
	}
	
	public String getName()
	{	return mName;
	}
	
	public void setMappings(IStandardEndpoint[] aMappings)
	{	mMappings=aMappings;
	}
	
	public Element getConfigElement()
	{	return mConfig;
	}
		
	public StandardSpace getParent()
	{	return mParentSpace;
	}
	
	public IKernel getKernel()
	{	return super.getKernel();
	}
	
	public String getOwnerName()
	{	return mOwner.getMeta().getIdentifier();
	}
	
	public StandardModule getOwningModule()
	{	return mOwner;
	}
	
	
	public int getIndexOfMapping(IStandardEndpoint aMapping)
	{	int result=0;
		if (mMappings!=null)
		{	for (int i=0; i<mMappings.length; i++)
			{	if (aMapping==mMappings[i])
				{	result=i;
					break;
				}
			}
		}
		return result;
	}	

	public ClassLoaderWithExports getClassLoader()
	{	return mOwner.getClassLoader();
	}

	public void onCommissionSpace(IKernel aKernel) throws Exception
	{	super.onCommissionSpace(aKernel);
	
		IKernel kernel=aKernel;
		int errors=0;
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.onCommissionEndpoint(kernel, this);
			}
			catch (Throwable e)
			{	mMappings[i]=null;
				errors++;
				String toString;
				ILogger logger=getKernel().getLogger();
				try
				{	toString = mapping.toString();
					if (toString==null)
					{	toString=mapping.getClass().getName();
					}
				} catch (Exception e2)
				{	toString=mapping.getClass().getName();
					logger.logRaw(ILogger.WARNING, this, Utils.throwableToString(e2));
				}
				String msg=logger.format("MSG_STD_ENDPOINT_COMMISSION_FAILED",mName,toString,Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
				setInitialisationProblem(msg);
			}
		}
		
		if (errors>0)
		{	IStandardEndpoint[] newMappings = new IStandardEndpoint[mMappings.length-errors];
			int j=0;
			for (int i=0; i<mMappings.length; i++)
			{	IStandardEndpoint mapping = mMappings[i];
				if (mapping!=null)
				{	newMappings[j++]=mapping;
				}
			}
			mMappings=newMappings;
		}	
		
	}
	
	void setInitialisationProblem(String aProblem)
	{	if (mProblems==RequestResponseFieldsImpl.EMPTY)
		{	mProblems = new RequestResponseFieldsImpl(10);
		}
		((RequestResponseFieldsImpl)mProblems).add(IEndpointStateMeta.KEY_PROBLEM, aProblem);
	}	
	
	public void postCommissionSpace() throws Exception
	{	super.postCommissionSpace();
		bindIfNeeded();
		Thread.currentThread().setContextClassLoader(getClassLoader());
        for (int i=0; i<mMappings.length; i++)
        {	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.postCommissionMapping(mStaticScope);
			}
			catch (Throwable e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_POSTCOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
        }
	}

	public void preDecommissionSpace() throws Exception
	{	Thread.currentThread().setContextClassLoader(getClassLoader());
		for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.preDecommissionMapping(mStaticScope);
			}
			catch (Exception e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_PREDECOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
		}		
		super.preDecommissionSpace();
	}	

	public void onDecommissionSpace() throws Exception
	{	for (int i=0; i<mMappings.length; i++)
		{	IStandardEndpoint mapping = mMappings[i];
			try
			{	mapping.onDecommissionEndpoint();
			}
			catch (Exception e)
			{	ILogger logger=getKernel().getLogger();
				String msg=logger.format("MSG_STD_ENDPOINT_DECOMMISSION_FAILED",mapping.toString(),Utils.throwableToString(e));
				logger.logRaw(ILogger.WARNING,this,msg);
			}
		}	
		super.onDecommissionSpace();
	}
	
	protected boolean isThisSpaceResolve(String aRequestId)
	{	return aRequestId.length()==0;
	}
	
	public void onResolve(INKFResolutionContext aContext) throws Exception
	{	
		if (aContext.isDebugResolve())
		{	onResolveDebug(aContext);
		}
		else
		{	INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
			if (toResolve.getVerb()==INKFRequestReadOnly.VERB_META)
			{	String idString=toResolve.getIdentifier();
				if (isThisSpaceResolve(idString))
				{	// resolve for space meta
					aContext.createResolutionResponse(this);
				}
				else
				{	StandardSpaceMeta meta = innerGetSpaceMeta(aContext);
					String absoluteIdentifier = meta.getAbsoluteIdentifier(idString);
					if (absoluteIdentifier !=null )
					{	IStandardEndpoint mapping=meta.getMappingWithId(absoluteIdentifier);
						IRequestScopeLevel parentScope = ((IRequest)aContext.getKernelContext().getThisKernelRequest().getValue(IRequest.TYPE_RESOLVE_REQUEST)).getRequestScope();
						aContext.createResolutionResponse((IEndpoint)mapping,absoluteIdentifier,getStaticScope(parentScope));
					}
					else
					{	aContext.createResponseFrom(null);
					}
				}
			}
			else
			{	//resolve for verbs other than META
				bindIfNeeded();
				for (int i=0; i<mMappings.length; i++)
				{	IStandardEndpoint mapping = mMappings[i];
					((NKFResolutionContextImpl)aContext).setEndpoint(mapping);
					mapping.onResolve(aContext);
					if (aContext.isResponseSet()) break;
				}
			}
		}
	}
	
	private void onResolveDebug(INKFResolutionContext aContext) throws Exception
	{
		INKFRequestReadOnly toResolve=aContext.getRequestToResolve();
		if (toResolve.getVerb()==INKFRequestReadOnly.VERB_META)
		{	String idString=toResolve.getIdentifier();
			if (isThisSpaceResolve(idString))
			{	// resolve for space meta
				aContext.createResolutionResponse(this);
				addResolveDebug(aContext,"STDR_CTX_META",null);
			}
			else
			{	StandardSpaceMeta meta = innerGetSpaceMeta(aContext);
				IStandardEndpoint mapping=meta.getMappingWithId(idString);
				if (mapping instanceof IEndpoint)
				{	aContext.createResolutionResponse((IEndpoint)mapping);
					addResolveDebug(aContext,"STDR_PART_META",mapping);
				}
			}
		}
		else
		{	//resolve for verbs other than META		
			for (int i=0; i<mMappings.length; i++)
			{	IStandardEndpoint mapping = mMappings[i];
				mapping.onResolve(aContext);
				if (aContext.isResponseSet())
				{	addResolveDebug(aContext,"STDR_MATCH",mapping);
					break;
				}
				else
				{	addResolveDebug(aContext,"STDR_NO_MATCH",mapping);
				}
			}
		}		
	}

	private void addResolveDebug(INKFResolutionContext aContext, String aMessageId, Object aOptionalParameter)
	{	String message;
		if (aOptionalParameter==null)
		{	message=getKernel().getLogger().format(aMessageId);
		}
		else
		{	message=getKernel().getLogger().format(aMessageId,aOptionalParameter.toString());
		}
		aContext.addResolutionDebug(message);
	}
	
	/** Append static scope on to end of a parent scope
	 * @param aParent
	 * @return
	 */
	private IRequestScopeLevel getStaticScope(IRequestScopeLevel aParent)
	{	IRequestScopeLevel scope = RequestScopeLevelImpl.createOrphanedRootScopeLevel(mStaticScopeArray[mStaticScopeArray.length-1], aParent);
		for (int i=mStaticScopeArray.length-2; i>=0; i--)
		{	scope = RequestScopeLevelImpl.createNonDurableScopeLevel(mStaticScopeArray[i], scope);
		}
		
		return scope;
	}
	
	private StandardSpaceMeta innerGetSpaceMeta(INKFRequestContext aContext) throws Exception
	{	
		INKFResponseReadOnly resp=mLastMetaResponse;
		if (resp==null || resp.isExpired())
		{	INKFRequest req=aContext.createRequest("");
			req.setVerb(INKFRequestReadOnly.VERB_META);
			req.setRepresentationClass(StandardSpaceMeta.class);
			resp = aContext.issueRequestForResponse(req);
			mLastMetaResponse=resp;
		}
		StandardSpaceMeta result = (StandardSpaceMeta)resp.getRepresentation();
		return result;
	}
	
	public void onMeta(INKFRequestContext aContext) throws Exception
	{	
		INKFResponseReadOnly resp=mLastMetaResponse;
		if (resp!=null && !resp.isExpired())
		{	INKFResponse respOut=aContext.createResponseFrom(resp.getRepresentation());
			aContext.getKernelContext().declareDependency(resp);
		}
		else
		{
		
			TupleList partsTuples = new TupleList(3,32);
			IRequestScopeLevel scope = getStaticScope(aContext.getKernelContext().getThisKernelRequest().getRequestScope());
			StandardSpaceMeta result = new StandardSpaceMeta(mName,null,null,null);
			
			
			bindIfNeeded();
			for (int i=0; i<mMappings.length; i++ )
			{	IStandardEndpoint mapping = mMappings[i];
				try
				{	((NKFContextImpl)aContext).setRequestScope(scope);
					((NKFContextImpl)aContext).setEndpoint(mapping);
					ISpaceElements parts = mapping.getParts(aContext);
					if (parts!=null)
					{	if (parts.size()>0)
						{	for (int j=0; j<parts.size(); j++)
							{	IIdentifier partId = parts.getIdentifier(j);
								if (partId!=null)
								{	result.putMapping(partId.toString(), mapping);
								}
								IEndpoint ep=parts.getPhysicalEndpoint(j);
								if (ep==null) ep=mapping;
								partsTuples.put(partId,parts.getDelegatedSpace(j),ep);
							}
						}
						else
						{	partsTuples.put(null,null,mapping);
						}
					}
				}
				catch (Exception e)
				{	INetKernelException ex = Utils.createFormattedException("EX_STD_DURING_GETPARTS", "MSG_STD_DURING_GETPARTS", getKernel().getLogger(), e, mapping.toString(),this.toString());
					String msg=ex.toString();
					getKernel().getLogger().logRaw(ILogger.WARNING, this, msg);
					setInitialisationProblem(msg);
				}
			}
			
			result.setParts(new SpaceElementsImpl(partsTuples));
			
			//state
			result.setAdditionalFields(mProblems);
			
			INKFResponse response =aContext.createResponseFrom(result);
			if (!mIsBound)
			{	response.setExpiry(INKFResponse.EXPIRY_MIN_FUNCTION_DEPENDENT,mUnboundExpiry);
				mLastMetaResponse=null;
			}
		}
	}
	
	private void bindIfNeeded() throws Exception
	{		
		//boolean bound=!mIsBinding;
		if (!mIsBound && !mIsBinding)
		{	mIsBinding=true;
			boolean bound=true;
			NKFLifecycleEventContextImpl context = null;
			try
			{		
				for (int i=0; i<mMappings.length; i++)
				{	IStandardEndpoint mapping = mMappings[i];
					if (mapping instanceof ILateBound)
					{	ILateBound lbp = (ILateBound)mapping;
						try
						{	if (!lbp.isBound())
							{	if (context==null)
								{	context = new NKFLifecycleEventContextImpl(getKernel(),mStaticScope,this,getParameters());
								}
								IStandardEndpoint boundMapping =lbp.bind(context);
								boundMapping.onCommissionEndpoint(getKernel(), this);
								mMappings[i]=boundMapping;
								mapping=mMappings[i];
								
								if (mUnboundExpiry!=null)
								{	mUnboundExpiry.invalidate();
									mUnboundExpiry=null;
								}
							}
						}
						catch (INetKernelException e)
						{	Element config=XMLUtils.getFirstChildElement(mConfig);
							for (int j=0; j<i && config!=null; j++)
							{	config=XMLUtils.getNextSiblingElement(config);
							}
							if (config==null) config=mConfig;
							SpaceFactory.logConstructionError(e, config, (StandardSpace)getSpace(),getKernel());
							bound=false;
						}
					}
				}
			}
			finally
			{	mIsBinding=false;
				mIsBound=bound;
			}
		}
		if (!mIsBound)
		{	if (mUnboundExpiry==null)
			{	mUnboundExpiry=new GoldenThreadExpiryFunction("StandardSpace unbound");
			}			
		}
	}
	
	public String toString()
	{	return mName;
	}
}